#!/usr/bin/python

import sys
import re
import csv

class FPGAPin:
    def __init__(self, name, bank, pad):
        self.name = name
        self.bank = int(bank)
        self.pad = pad
        self.in_use = False

def differ_by_single_letter(s1, s2):
    if len(s1) != len(s2):
        return False;

    ch = 0
    last = None
    for i in range(0, len(s1)):
        if s1[i] != s2[i]:
            ch += 1
            last = i
    if ch == 1:
        return ( s1[last], s2[last] )
    else:
        return False

class FPGAPinFile:
    def __init__(self, filename):
        self.pins = []
        for l in open(filename).readlines():
            tok = l.split()
            if len(tok) < 8:
                continue

            if tok[6] == "HR" or tok[6] == "HP" or tok[6] == "GTX":
                self.pins += [ FPGAPin(tok[1], tok[3], tok[0])]

    def find_pin(self, name, banks=None):
        rv = None
        for p in self.pins:
            if banks and not p.bank in banks:
                continue
            if p.name == name:
                return p
        return None

    def first_free_pin(self, banks=None):
        rv = None
        for p in self.pins:
            if p.in_use:
                continue
            if banks and not p.bank in banks:
                continue
            return p

    def first_free_diff_pair(self, banks=None):
        rv = None
        
        for p_ref in self.pins:
            if p_ref.in_use:
                continue
            id1 = re.match("IO_L(\d+)([PN]).+", p_ref.name)
            if id1 == None:
                continue
            
            p_ref_n = id1.group(1)
            p_ref_pol = id1.group(2)
            
            for p in self.pins:
                if p.in_use:
                    continue
                if banks and not p.bank in banks:
                    continue
                if p.bank != p_ref.bank:
                    continue
                id2 = re.match("IO_L(\d+)([PN]).+", p.name)
                if id2 == None:
                    continue


                p_n = id2.group(1)
                p_pol = id2.group(2)
                
                if p_ref_n == p_n and p_ref_pol != p_pol:
                #    print("MAtch!")
                    if p_ref_pol.lower() == 'p':
                        return ( p_ref, p )
                    else:
                        return ( p, p_ref )

        return rv

    def free_pins_in_banks(self, banks):
        for b in banks:
            cnt_free = 0
            cnt_total = 0
            for p in self.pins:
                if p.bank == b:
                    cnt_total += 1
                    if not p.in_use:
                        cnt_free += 1
            print("Bank %d: %d total, %d free pins" % (b, cnt_total, cnt_free))

class Signal: 
    def __init__(self, name, range_hi = 0, range_lo = 0, diff=False, force=None, banks=None, iostandard=None):
        self.name = name
        self.range_hi = range_hi
        self.range_lo = range_lo
        self.diff = diff
        self.force = force
        self.banks = banks
        self.iostandard = iostandard
        self.done=False
        self.pair=None
        self.fpga_pins={}
        

class TopLevelSignals:
    def __init__(self):
        self.signals=[]

    def add( self, name, range_hi = 0, range_lo = 0, diff=False, force=None, banks=None, iostandard=None):
        self.signals += [ Signal( name, range_hi, range_lo, diff, force, banks, iostandard)]

    def print_stats(self):
        total = 0
        for s in self.signals:
            total += (s.range_hi + 1 - s.range_lo)

        print("Total signals : %d", total)

    def find_diff_pair(self, p_ref):
        rv = None
        
        for p in self.signals:
            ch = 0
            #print("%s %s", p.name, p_ref.name)
            if len(p.name) != len(p_ref.name):
                continue
            for i in range(0, len(p_ref.name)):
                #print ( p.name[i].lower() )
                if p.name[i].lower() != p_ref.name[i].lower():
                    ch+=1
                if p.name[i].lower()=="p" and p_ref.name[i].lower()=="n":
                    #print("MATCH")
                    rv = (p, p_ref)
                elif p.name[i].lower()=="n" and p_ref.name[i].lower()=="p":
                    #print("MATCH2")
                    rv = (p_ref, p)
            if ch == 1 and rv:
                return rv

        return rv

fpga_pins = FPGAPinFile( 'xc7k325tffg900pkg.txt')

signals = TopLevelSignals()

banks_hp = [32, 33, 34]
banks_hr_33 = [12, 13, 17]
banks_hr_25 = [14, 15, 16, 18]

banks_hr_all = [12, 13, 14, 15, 16, 17, 18]

signals.add("clk_20m_vcxo_i", force="IO_L13P_T2_MRCC_13")
signals.add("clk_62m5_pllref_p_i", force="IO_L12P_T1_MRCC_12", iostandard="LVDS")
signals.add("clk_62m5_pllref_n_i", force="IO_L12N_T1_MRCC_12", iostandard="LVDS")
signals.add("clk_125m_gtx_n_i", force="MGTREFCLK0N_115")
signals.add("clk_125m_gtx_p_i", force="MGTREFCLK0P_115")
signals.add("clk_fpga2_p_i", force="IO_L13P_T2_MRCC_12", iostandard="LVDS")
signals.add("clk_fpga2_n_i", force="IO_L13N_T2_MRCC_12", iostandard="LVDS")
signals.add("clk_si57x_p_i", force="IO_L12P_T1_MRCC_13", iostandard="LVDS")
signals.add("clk_si57x_n_i", force="IO_L12N_T1_MRCC_13", iostandard="LVDS")

signals.add("fmc0_clk_m2c_p_i[0]", force="IO_L12P_T1_MRCC_AD5P_15", banks=banks_hr_25, iostandard="LVDS_25")
signals.add("fmc0_clk_m2c_n_i[0]", force="IO_L12N_T1_MRCC_AD5N_15", banks=banks_hr_25, iostandard="LVDS_25")
signals.add("fmc0_clk_m2c_p_i[1]", force="IO_L13P_T2_MRCC_15", banks=banks_hr_25, iostandard="LVDS_25")
signals.add("fmc0_clk_m2c_n_i[1]", force="IO_L13N_T2_MRCC_15", banks=banks_hr_25, iostandard="LVDS_25")

signals.add("fmc1_clk_m2c_p_i[0]", force="IO_L12P_T1_MRCC_16", banks=banks_hr_25, iostandard="LVDS_25")
signals.add("fmc1_clk_m2c_n_i[0]", force="IO_L12N_T1_MRCC_16", banks=banks_hr_25, iostandard="LVDS_25")
signals.add("fmc1_clk_m2c_p_i[1]", force="IO_L13P_T2_MRCC_16", banks=banks_hr_25, iostandard="LVDS_25")
signals.add("fmc1_clk_m2c_n_i[1]", force="IO_L13P_T2_MRCC_16", banks=banks_hr_25, iostandard="LVDS_25")

signals.add("fmc0_la_p_b", 33, 0, diff=True, banks=banks_hr_25)
signals.add("fmc0_la_n_b", 33, 0, diff=True, banks=banks_hr_25)
signals.add("fmc1_la_p_b", 33, 0, diff=True, banks=banks_hr_25)
signals.add("fmc1_la_n_b", 33, 0, diff=True, banks=banks_hr_25)

signals.add("p2_p_b", 9, 0, diff=True, banks=banks_hr_33)
signals.add("p2_n_b", 9, 0, diff=True, banks=banks_hr_33)

signals.add("p2_p_b", 19, 10, diff=True, banks=banks_hr_25)
signals.add("p2_n_b", 19, 10, diff=True, banks=banks_hr_25)


signals.add("rst_n_i", banks = banks_hr_33)
signals.add("vme_write_n_i", banks = banks_hr_33)
signals.add("vme_sysreset_n_i", banks = banks_hr_33)
signals.add("vme_retry_oe_o", banks = banks_hr_33)
signals.add("vme_retry_n_o", banks = banks_hr_33)
signals.add("vme_lword_n_b", banks = banks_hr_33)
signals.add("vme_iackout_n_o", banks = banks_hr_33)
signals.add("vme_iackin_n_i", banks = banks_hr_33)
signals.add("vme_iack_n_i", banks = banks_hr_33)
signals.add("vme_gap_i", banks = banks_hr_33)
signals.add("vme_dtack_oe_o", banks = banks_hr_33)
signals.add("vme_dtack_n_o", banks = banks_hr_33)
signals.add("vme_ds_n_i", 1, 0, banks = banks_hr_33)
signals.add("vme_data_oe_n_o", banks = banks_hr_33)
signals.add("vme_data_dir_o", banks = banks_hr_33)
signals.add("vme_berr_o", banks = banks_hr_33)
signals.add("vme_as_n_i", banks = banks_hr_33)
signals.add("vme_addr_oe_n_o", banks = banks_hr_33)
signals.add("vme_addr_dir_o", banks = banks_hr_33)
signals.add("vme_irq_o",        7, 1, banks = banks_hr_33)
signals.add("vme_ga_i",         4, 0, banks = banks_hr_33)
signals.add("vme_data_b",       31, 0, banks = banks_hr_33)
signals.add("vme_am_i",         5, 0, banks = banks_hr_33)
signals.add("vme_addr_b",       31, 1, banks = banks_hr_33)
signals.add("vme_noga_i",       4, 0, banks = banks_hp)
signals.add("vme_use_ga_i", banks = banks_hp)
signals.add("pll20dac_din_o", banks=banks_hr_all),
signals.add("pll20dac_sclk_o", banks=banks_hr_all),
signals.add("pll20dac_sync_n_o", banks=banks_hr_all),
signals.add("pll25dac_din_o", banks=banks_hr_all),
signals.add("pll25dac_sclk_o", banks=banks_hr_all),
signals.add("pll25dac_sync_n_o", banks=banks_hr_all),
signals.add("sfp_txp_o", force="MGTXTXP0_115")
signals.add("sfp_txn_o", force="MGTXTXN0_115")
signals.add("sfp_rxp_i", force="MGTXRXP0_115")
signals.add("sfp_rxn_i", force="MGTXRXN0_115")
signals.add("sfp_mod_def0_i", banks=banks_hp),
signals.add("sfp_mod_def1_b", banks=banks_hr_all),
signals.add("sfp_mod_def2_b", banks=banks_hr_all),
signals.add("sfp_rate_select_o", banks=banks_hr_all),
signals.add("sfp_tx_fault_i", banks=banks_hp),
signals.add("sfp_tx_disable_o", banks=banks_hr_all),
signals.add("sfp_los_i", banks=banks_hp),
signals.add("carrier_scl_b", banks=banks_hr_all),
signals.add("carrier_sda_b", banks=banks_hr_all),
signals.add("pcbrev_i", 4, 0, banks=banks_hp),
signals.add("onewire_b", banks=banks_hr_all),
signals.add("uart_rxd_i", banks=banks_hr_all),
signals.add("uart_txd_o", banks=banks_hr_all),
signals.add("spi_ncs_o", banks=banks_hr_all),
signals.add("spi_mosi_o", banks=banks_hr_all),
signals.add("spi_miso_i", banks=banks_hr_all),

signals.add("fp_led_line_oen_o", 1,0, banks=banks_hr_all)
signals.add("fp_led_line_o", 1,0, banks=banks_hr_all)
signals.add("fp_led_column_o",   3,0, banks=banks_hr_all)

signals.add("fp_gpio1_b", banks=banks_hr_all),
signals.add("fp_gpio2_b", banks=banks_hr_all),
signals.add("fp_gpio3_b", banks=banks_hr_all),
signals.add("fp_gpio4_b", banks=banks_hr_all),
signals.add("fp_term_en_o",4, 1, banks=banks_hp)
signals.add("fp_gpio1_a2b_o", banks=banks_hr_all)
signals.add("fp_gpio2_a2b_o", banks=banks_hr_all)
signals.add("fp_gpio34_a2b_o", banks=banks_hr_all)

signals.add("fp_pushbutton_n_i", banks=banks_hp),

signals.add("fmc0_pg_i", banks=banks_hr_25)
signals.add("fmc0_prsnt_m2c_n_i", banks=banks_hr_25)

signals.add("fmc0_scl_b", banks=banks_hr_25)
signals.add("fmc0_sda_b", banks=banks_hr_25)
signals.add("fmc0_tck_o", banks=banks_hr_25)
signals.add("fmc0_tms_o", banks=banks_hr_25)
signals.add("fmc0_tdi_o", banks=banks_hr_25)
signals.add("fmc0_tdo_i", banks=banks_hr_25)


signals.add("fmc1_pg_i", banks=banks_hp)
signals.add("fmc1_prsnt_m2c_n_i", banks=banks_hp)

signals.add("fmc1_scl_b", banks=banks_hr_25)
signals.add("fmc1_sda_b", banks=banks_hr_25)
signals.add("fmc1_tck_o", banks=banks_hr_25)
signals.add("fmc1_tms_o", banks=banks_hr_25)
signals.add("fmc1_tdi_o", banks=banks_hr_25)
signals.add("fmc1_tdo_i", banks=banks_hr_25)

signals.add("ddr_a_o",   14, 0, banks=banks_hp)
signals.add("ddr_ba_o",  2, 0, banks=banks_hp)
signals.add("ddr_cas_n_o", banks=banks_hp)
signals.add("ddr_ck_n_o",diff=True, banks=banks_hp)
signals.add("ddr_ck_p_o",diff=True, banks=banks_hp)
signals.add("ddr_cke_o", banks=banks_hp)
signals.add("ddr_dq_b",  63, 0, banks=banks_hp)
signals.add("ddr_dqs_n_b",   7, 0, diff=True, banks=banks_hp)
signals.add("ddr_dqs_p_b",   7, 0, diff=True, banks=banks_hp)
signals.add("ddr_odt_o", banks=banks_hp)
signals.add("ddr_ras_n_o", banks=banks_hp)
signals.add("ddr_reset_n_o", banks=banks_hp)
signals.add("ddr_rzq_b", banks=banks_hp)
signals.add("ddr_dm_o",  7, 0, banks=banks_hp)
signals.add("ddr_we_n_o", banks=banks_hp)
signals.add("ddr_cs_n_o", banks=banks_hp)
signals.add("si57x_scl_b", banks=banks_hr_all)
signals.add("si57x_sda_b", banks=banks_hr_all)
signals.add("si57x_oe_o", banks=banks_hr_all)
signals.add("si57x_tune_o", banks=banks_hr_all)

for s in signals.signals:
    if s.done:
        continue
    if s.force:
        p = fpga_pins.find_pin(s.force)
        if p == None:
            print("!!!! unknown forced pin")
        print("Force: %-30s -> %-20s" % (s.name, s.force))
        s.done = True
        s.fpga_pins[0] = p



for s in signals.signals:
    if s.done:
        continue
    if s.diff:
        for idx in range( s.range_lo, s.range_hi + 1 ):
            #print("Do Signal %s[%d]" % (s.name, idx))
            pair = signals.find_diff_pair( s )
            f_pair = fpga_pins.first_free_diff_pair( banks = s.banks )
            if f_pair == None:
                print("!!!! Unable to fit diff pair: %s" % s.name)
                break
            pair[0].done = True
            pair[1].done = True
            f_pair[0].in_use = True
            f_pair[1].in_use = True

    
            pair[0].fpga_pins[idx] = f_pair[0]
            pair[1].fpga_pins[idx] = f_pair[1]

            #print("DP %s[%d] %s %s %d" %( pair[0].name, idx, f_pair[0].name, f_pair[0].pad, f_pair[0].bank ) )
            #print("DN %s[%d] %s %s %d" %( pair[1].name, idx, f_pair[1].name, f_pair[1].pad, f_pair[1].bank ) )
    else:
        for idx in range( s.range_lo, s.range_hi + 1 ):
            f_pin = fpga_pins.first_free_pin( banks = s.banks )
            if f_pin == None:
                print("!!!! Unable to fit pin: %s" % s.name)
                break
            s.done = True
            f_pin.in_use = True
            s.fpga_pins[idx] = f_pin

            #print("D %s[%d] %s %s %d" %( s.name, idx, f_pin.name, f_pin.pad, f_pin.bank ) )

f_xdc = open("out.xdc","wb")

for s in signals.signals:
    for idx in range( s.range_lo, s.range_hi + 1 ):
        if s.range_lo == 0 and s.range_hi == 0:
            rstring = ""
        else:
            rstring = "[%d]" % idx
        f_xdc.write("set_property PACKAGE_PIN %s [get_ports {%s%s}]\n" % (s.fpga_pins[idx].pad, s.name, rstring))
        if not s.diff:
            iostd = None
            if s.iostandard != None:
                iostd = s.iostandard
            elif s.fpga_pins[idx].bank in banks_hr_33:
                iostd = "LVCMOS33"
            elif s.fpga_pins[idx].bank in banks_hr_25:
                iostd = "LVCMOS25"
            if iostd:
                f_xdc.write("set_property IOSTANDARD %s [get_ports {%s%s}]\n"% (iostd, s.name, rstring))

fpga_pins.free_pins_in_banks([12,13,14,15,16,17,18,32,33,34])
signals.print_stats()

f_xdc.close()